#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "dcpu16.h"
#include "hw_clock.h"

static dcpu16_hw_signal_t clock_reset_;
static dcpu16_hw_signal_t clock_cycle_;
static dcpu16_hw_signal_t clock_hwi_;
static struct dcpu16_hw hw_ = {
    .name_  = "Generic Clock (compatible)",
    .id_l   = 0xb402,
    .id_h   = 0x12d0,
    .ver    = 0x0001,
    .mfg_l  = 0x0000,
    .mfg_h  = 0x0000,
    .hwi    = clock_hwi_,
    .cycle  = clock_cycle_,
    .reset  = clock_reset_,
    .data   = (struct clock_ *)NULL
};

struct clock_ {
    DCPU16_WORD cycle_;
    DCPU16_WORD rate;
    DCPU16_WORD tick;
    DCPU16_WORD interrupt_message;
};

static
void clock_reset_(struct dcpu16 *vm, void *data) {
    struct clock_ *clock = (struct clock_ *)data;

    (void)vm;

    memset(clock, 0, sizeof *clock);
}

static
void clock_cycle_(struct dcpu16 *vm, void *data) {
    struct clock_ *clock = (struct clock_ *)data;

    /* cycle is only called 100000 times per second */
    /* maximum rate is 60hz / word_max = 3932160 */

    if (clock->rate == 0)
        return;

    clock->cycle_++;
    if (clock->cycle_ >= clock->rate) {
        /* THIS CHECK IS WRONG, JUST A PLACEHOLDER */
        clock->cycle_ = 0;
        clock->tick += 1;

        if (clock->interrupt_message) {
            if (dcpu16_interrupt(vm, clock->interrupt_message))
                vm->warn_cb_("%s: could not send interrupt", hw_.name_);
        }
    }
}

static
void clock_hwi_(struct dcpu16 *vm, void *data) {
    struct clock_ *clock = (struct clock_ *)data;
    DCPU16_WORD reg_a = vm->reg[DCPU16_REG_A];
    DCPU16_WORD reg_b = vm->reg[DCPU16_REG_B];

    switch (reg_a) {
        case 0: /* set tick gather rate, 60hz/B */
            clock->rate = reg_b;
            break;

        case 1: /* fetch elapsed count since rate was set */
            vm->reg[DCPU16_REG_C] = clock->tick;
            break;

        case 2:
            clock->interrupt_message = reg_b;
            break;
    }
}

/* instantitate a new clock */
struct dcpu16_hw *clock_new(struct dcpu16 *vm) {
    struct dcpu16_hw *hw;

    hw = calloc(1, sizeof *hw);
    if (hw == NULL) {
        vm->warn_cb_("%s():%s", "calloc", strerror(errno));
        return NULL;
    }
    memcpy(hw, &hw_, sizeof *hw);
    hw->data = calloc(1, sizeof hw->data);
    if (hw->data == NULL) {
        vm->warn_cb_("%s():%s", "calloc", strerror(errno));
        free(hw);
        return NULL;
    }

    return hw;
}

void clock_del(struct dcpu16_hw **hw) {
    if (hw) {
        free((*hw)->data);
        (*hw)->data = NULL;

        free((*hw));
        *hw = NULL;
    }
}
